
module grace_cstk;

import grace_def;
import grace_cstk;
import grace_dict;
import grace_eval;
import grace_reader;

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// -------------------- globals

static const auto typeCheck=false;

const int CT_INT =  1; // integer
const int CT_FLT =  2; // float
const int CT_SYM =  3; // symbol
const int CT_STR =  4; // string
const int CT_HDR =  5; // header
const int CT_LST =  6; // list
const int CT_PTR =  7; // pointer
const int CT_NFP =  8; // native function pointer
const int CT_LOG =  9; // logical
const int CT_REX = 10; // regular expression
const int CT_BAR = 11; // byte array
const int CT_ASA = 12; // associative array
const int CT_CEL = 13; // associative array
const int CT_CHR = 14; // character

int TH_INT; // integer
int TH_FLT; // float
int TH_SYM; // symbol
int TH_STR; // string
int TH_HDR; // header
int TH_LST; // list
int TH_PTR; // pointer
int TH_NFP; // native funUHion pointer
int TH_LOG; // logical
int TH_REX; // regular expression
int TH_BAR; // byte array
int TH_ASA; // associative array
int TH_CEL; // cell
int TH_CHR; // char

struct symbol
{
  char[]  s; // symbol
  header *c; // context
}

struct cell
{
  union
  {
    int      i;
    float    f;
    nfun     n;
    void    *p;
    header  *h;
//    char[]  *s;
    Box     *b;
    symbol  *y;
    char     c;
  };
//  int    tc; // type code
  int    th; // type hash
//  char[] ts; // type string
}

cell[]   stk,rst;
cell[][] stkstk,rststk;

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// -------------------- cells

void types_init()
{
  str2uh("nul");
  TH_INT=str2uh("int"); // integer
  TH_FLT=str2uh("flt"); // float
  TH_SYM=str2uh("sym"); // symbol
  TH_STR=str2uh("str"); // string
  TH_HDR=str2uh("hdr"); // header
  TH_LST=str2uh("lst"); // list
  TH_PTR=str2uh("ptr"); // pointer
  TH_NFP=str2uh("nfp"); // native function pointer
  TH_LOG=str2uh("log"); // logical
  TH_REX=str2uh("rex"); // regular expression
  TH_BAR=str2uh("bar"); // byte array
  TH_ASA=str2uh("asa"); // associative array
  TH_CEL=str2uh("cel"); // cell
  TH_CHR=str2uh("chr"); // char
}

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// -------------------- cell conversion

cell int2cell(int v)
{
  cell c;
  c.i=v;
//  c.tc=CT_INT;
  c.th=TH_INT;
//  c.ts="int";
  return c;
}

int cell2int(cell c)
{
  static if (typeCheck)
  {
    if (c.th!=TH_INT)
      throw new Exception("wrong type, INT expected");
  }
  return c.i;
}

cell chr2cell(char v)
{
  cell c;
  c.i=v;
//  c.tc=CT_INT;
  c.th=TH_CHR;
//  c.ts="int";
  return c;
}

char cell2chr(cell c)
{
  static if (typeCheck)
  {
    if (c.th!=TH_CHR)
      throw new Exception("wrong type, CHR expected");
  }
  return c.c;
}

cell log2cell(int v)
{
  cell c;
//  c.tc=CT_LOG;
  c.th=TH_LOG;
//  c.ts="log";
  c.i=v;
  return c;
}

int cell2log(cell c)
{
  static if (typeCheck)
  {
    if (c.th!=TH_LOG)
      throw new Exception("wrong type, LOG expected");
  }
  return c.i;
}

cell flt2cell(float v)
{
  cell c;
//  c.tc=CT_FLT;
  c.th=TH_FLT;
//  c.ts="flt";
  c.f=v;
  return c;
}

float cell2flt(cell c)
{
  static if (typeCheck)
  {
    if (c.th!=TH_FLT)
      throw new Exception("wrong type, FLT expected");
  }
  return c.f;
}

cell hdr2cell(header* v)
{
  cell c;
//  c.tc=CT_HDR;
  c.th=TH_HDR;
//  c.ts="hdr";
  c.p=cast(void*)v;
  return c;
}

header* cell2hdr(cell c)
{
  static if (typeCheck)
  {
    if (c.th!=TH_HDR)
      throw new Exception("wrong type, HDR expected");
  }
  return cast(header*)c.p;
}

cell str2cell(char[] v)
{
  cell c;
  Box *b=new Box;
  *b=box(v);
//  c.tc=CT_STR;
  c.th=TH_STR;
//  c.ts="str";
  c.b=b;
  return c;
}

char[] cell2str(cell c)
{
  static if (typeCheck)
  {
    if (c.th!=TH_STR)
      throw new Exception("wrong type, STR expected");
  }
  return unbox!(char[])(*(c.b));
}

cell sym2cell(symbol *y)
{
  cell c;
//  c.tc=CT_SYM;
  c.th=TH_SYM;
//  c.ts="sym";
  c.y=y;
  return c;
}

cell sym2cell(char[] s,header *c)
{
  symbol *y=new symbol;
  y.s=s;
  y.c=c;
  return sym2cell(y);
}

symbol cell2sym(cell c)
{
  static if (typeCheck)
  {
    if (c.th!=TH_SYM)
      throw new Exception("wrong type, SYM expected");
  }
  return *(c.y);
}

cell lst2cell(cell[] v)
{
  cell c;
  Box *b=new Box;
  *b=box(v);
//  c.tc=CT_LST;
  c.th=TH_LST;
//  c.ts="lst";
  c.b=b;
  return c;
}

cell[] cell2lst(cell c)
{
  static if (typeCheck)
  {
    if (c.th!=TH_LST)
      throw new Exception("wrong type, LST expected");
  }
  return unbox!(cell[])(*(c.b));
}

cell ptr2cell(void* v)
{
  cell c;
//  c.tc=CT_PTR;
  c.th=TH_PTR;
//  c.ts="ptr";
  c.p=v;
  return c;
}

void* cell2ptr(cell c)
{
  static if (typeCheck)
  {
    if (c.th!=TH_PTR)
      throw new Exception("wrong type, PTR expected");
  }
  return c.p;
}

cell cel2cell(cell v)
{
  cell c;
  Box *b=new Box;
  *b=box(v);
  c.th=TH_CEL;
  c.b=b;
  return c;
}

cell cell2cel(cell c)
{
  static if (typeCheck)
  {
    if (c.th!=TH_CEL)
      throw new Exception("wrong type, CEL expected");
  }
  return unbox!(cell)(*(c.b));
}

cell nfp2cell(void* v)
{
  cell c;
//  c.tc=CT_NFP;
  c.th=TH_NFP;
//  c.ts="nfp";
  c.p=v;
  return c;
}

nfun cell2nfp(cell c)
{
  static if (typeCheck)
  {
    if (c.th!=TH_NFP)
      throw new Exception("wrong type, NFP expected");
  }
  return c.n;
}

cell rex2cell(RegExp *v)
{
  cell c;
//  c.tc=CT_REX;
  c.th=TH_REX;
//  c.ts="rex";
  c.p=cast(void*)v;
  return c;
}

RegExp cell2rex(cell c)
{
  static if (typeCheck)
  {
    if (c.th!=TH_REX)
      throw new Exception("wrong type, REX expected");
  }
  return *(cast(RegExp*)c.p);
}

cell bar2cell(void[] v)
{
  cell c;
  Box *b=new Box;
  *b=box(v);
//  c.tc=CT_BAR;
  c.th=TH_BAR;
//  c.ts="bar";
  c.b=b;
  return c;
}

void[] cell2bar(cell c)
{
  static if (typeCheck)
  {
    if (c.th!=TH_BAR)
      throw new Exception("wrong type, BAR expected");
  }
  return unbox!(void[])(*(c.b));
}

cell asa2cell(cell[char[]] v)
{
  cell c;
  Box *b=new Box;
  *b=box(v);
//  c.tc=CT_ASA;
  c.th=TH_ASA;
//  c.ts="asa";
  c.b=b;
  return c;
}

cell[char[]] cell2asa(cell c)
{
  static if (typeCheck)
  {
    if (c.th!=TH_ASA)
      throw new Exception("wrong type, ASA expected");
  }
  return unbox!(cell[char[]])(*(c.b));
}

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// -------------------- cell string visualisation

char[] lst_vs(cell[] l)
{
  char[] s;
  int k;
  s="[";
  if (l.length)
    s~=cell_vs(l[k]);
  for (k=1;k<l.length;k++)
    s~=" "~cell_vs(l[k]);
  s~="]";
  return s;
}

char[] xxx(char[] p)
{
  char[] s=p.dup;
  for (int k=0;k<s.length;k++)
  {
    if (cast(int)s[k]<32)
      s[k]='.';
  }
  return s;
}

char[] cell_vs(cell c)
{
  char[] s;
/*
  writef("cell_vs(%s",uh2str(c.th));
  if (c.th==CT_NFP)
  {
    writef(" NFP");
    printf(" %p",c.p);
  }
*/
  switch (c.th)
  {
    case CT_INT : s=swritef(cell2int(c));break;
    case CT_FLT : s=swritef(cell2flt(c));break;
//    case CT_STR : s=swritef("STR@%p=\"",c.p,"%s",xxx(cell2str(c)),"\"");break;
    case CT_STR : s="\""~cell2str(c)~"\"";break;
    case CT_CHR : s=""~cell2chr(c);break;
    case CT_SYM : s=cell2sym(c).s;break;
    case CT_HDR : s=swritef("HDR@%p",c.p);break;
    case CT_NFP : s="NFP@"~swritef(c.p);break;
//    case CT_LST : sprintf(s,"LST[%p]",c.p);break;
    case CT_LST : s=swritef("%s",lst_vs(cell2lst(c)));break;
    case CT_BAR : s=swritef("BAR@%p",c.p);break;
    case CT_ASA : s=swritef("ASA@%p",c.p);break;
    case CT_LOG : {if (cell2log(c)) s="#t\0"; else s="#f\0";break;}
    default     : s=swritef("?@%p",c.p);
  }
//  writefln(")");
  return s;
}

char[] cell_vs1(cell c)
{
  char[] s;
//  writef("cell_vs(%s",uh2str(c.th));
  switch (c.th)
  {
    case CT_INT : s=swritef(cell2int(c));break;
    case CT_FLT : s=swritef(cell2flt(c));break;
//    case CT_STR : s=swritef("STR@%p=\"",c.p,"%s",xxx(cell2str(c)),"\"");break;
    case CT_STR : s=swritef("%s",xxx(cell2str(c)));break;
    case CT_CHR : s=swritef("'%c'",cell2chr(c));break;
    case CT_SYM : s=swritef("SYM@%p=%s",c.p,xxx(cell2sym(c).s));break;
    case CT_HDR : s=swritef("HDR@%p",c.p);break;
    case CT_NFP : s="NFP@"~swritef("%p",c.p);break;
//    case CT_LST : sprintf(s,"LST[%p]",c.p);break;
    case CT_LST : s=swritef("LST@%p=%s",c.p,lst_vs(cell2lst(c)));break;
    case CT_BAR : s=swritef("BAR@%p",c.p);break;
    case CT_ASA : s=swritef("ASA@%p",c.p);break;
    case CT_LOG : {if (cell2log(c)) s="#t\0"; else s="#f\0";break;}
    default     : s=swritef("?@%p",c.p);
  }
//  writefln(")");
  return s;
}

cell[] new_lst()
{
  cell[] l;
  l.length=0x10;
  l.length=0;
  return l;
}

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// -------------------- stack

void stack_zapp()
{
  stk.length=0;
  rst.length=0;
}

void stack_top()
{
  rst=dpop()~rst;
}

int stack_depth()
{
  return stk.length;
}

int stack_height()
{
  return rst.length;
}

int dthat(int k)
{
  return stk[$-1-k].th;
}

cell dtos()
{
  if (stk.length)
    return (stk[$-1]);
  throw new Exception("dtos: no tos");
//  return cast(cell*)0;
}

void dtoset(cell c)
{
  if (stk.length)
  {
    stk[$-1]=c;
    return;
  }
  throw new Exception("dtoset: no tos");
}

void dpush(cell c)
// push cell
{
/*  if (!(dss[$-1].length&0xff))
  {
    dss[$-1].length=dss[$-1].length+0x100;
    dss[$-1].length=dss[$-1].length-0x100;
  }*/
  stk~=c;
}

cell dpop()
// pop cell
{
  if (!(stk.length))
  {
    src_show();
    throw new Exception("stack underflow");
  }
  cell c=stk[$-1];
  stk.length=stk.length-1;
  return c;
}

void denternew()
{
  dpush_lst(new_lst());
  denter();
}

void denter()
{
  cell[] l=dpop_lst();
  stkstk~=stk.dup;
  rststk~=rst.dup;
  stk=l.dup;
  rst.length=0;
}

void dleave()
{
  cell[] l=stk~rst;
  if (stkstk.length)
  {
    stk=stkstk[$-1].dup;
    stkstk.length=stkstk.length-1;
    rst=rststk[$-1].dup;
    rststk.length=rststk.length-1;
    dpush_lst(l.dup);
  }
  else
  {
    stk.length=0;
    rst.length=0;
    dpush_lst(l.dup);
  }
}

void dive()
{
  if (stk.length)
  {
    rst~=stk[$-1];
    stk.length=stk.length-1;
  }
}

void soar()
{
  if (rst.length)
  {
    stk~=rst[$-1];
    rst.length=rst.length-1;
  }
}

void xdive()
{
  while (stk.length)
  {
    rst~=stk[$-1];
    stk.length=stk.length-1;
  }
}

void xsoar()
{
  while (rst.length)
  {
    stk~=rst[$-1];
    rst.length=rst.length-1;
  }
}

void ds_show()
{
  cell c;
  int k;
//  printf("ds @%p: ",&dss[$-1]);
  printf("ds: ");
  try for (k=0;k<stk.length;k++)
  {
    c=(stk[k]);
    writef("%s   ",cell_vs(c));
  }
  catch
  {
    printf("ARGH!\n");
  }
  printf("\n");
}

void dss_show()
{
  cell c;
  int k,kk;
//  printf("dss@%p: ",&dss);
  printf("dss: ");
  try
  {
    for (kk=0;kk<stkstk.length;kk++)
    {
//      printf("LST@%p[ ",&dss[kk]);
      printf("[ ",&stkstk[kk]);
      for (k=0;k<stkstk[kk].length;k++)
      {
        c=(stkstk[kk][k]);
        writef("%s   ",cell_vs(c));
      }
      printf("] ");
    }
  }
  catch
  {
    printf("ARGH!\n");
  }
  printf("\n");
}

void dss_show_ll()
{
  int *p=cast(int*)&stkstk;
  int l,k;
  printf("stkstk=[%08x,%08x] -> ",p[0],p[1]);
  l=p[0]*2;
  p=cast(int*)p[1];
  for (k=0;k<l;k++)
    printf("%08x ",p[k]);
  printf("\n");
}

void ds_tst()
{
  int k;
  cell c;
//  c.tc=0;
  c.th=0;
//  c.ts="";
  printf("\nlen=%i\n",stk.length);
  for (k=0;k<0x1000;k++)
  {
    c.i=k;
    dpush(c);
  }
  while (stk.length)
  {
    c=dpop();
    printf("%i ",c.i);
  }
  printf("\nlen=%i\n",stk.length);
}

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// -------------------- typed stack functions (for convenience)

void dpush_asa(cell[char[]] v)
{
  dpush(asa2cell(v));
}

cell[char[]] dpop_asa()
{
  return cell2asa(dpop());
}

void dpush_bar(void[] v)
{
  dpush(bar2cell(v));
}

void[] dpop_bar()
{
  return cell2bar(dpop());
}

void dpush_int(int v)
{
  dpush(int2cell(v));
}

int dpop_int()
{
  return cell2int(dpop());
}

void dpush_chr(char v)
{
  dpush(chr2cell(v));
}

char dpop_chr()
{
  return cell2chr(dpop());
}

void dpush_log(int v)
{
  dpush(log2cell(v));
}

int dpop_log()
{
  return cell2log(dpop());
}

void dpush_flt(float v)
{
  dpush(flt2cell(v));
}

float dpop_flt()
{
  return cell2flt(dpop());
}

void dpush_str(char[] v)
{
  dpush(str2cell(v));
}

char[] dpop_str()
{
  return cell2str(dpop());
}

void dpush_sym(symbol v)
{
  dpush(sym2cell(&v));
}

symbol dpop_sym()
{
  return cell2sym(dpop());
}

void dpush_lst(cell[] v)
{
  dpush(lst2cell(v));
}

cell[] dpop_lst()
{
  return cell2lst(dpop());
}
