
module grace_eval;

import grace_def;
import grace_cstk;
import grace_dict;
import grace_reader;
import std.conv;

import grace_prim_system;

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// -------------------- eval / quote

void eval_hdr(header *h)
{
  if (!h)
  {
    throw new Exception("eval_hdr: null header");
  }
  int f=hdr_get_flg(h);
  if (f&FLG_F)
    eval(hdr_get_dat(h));
  else
    dpush(hdr_get_dat(h));
}

void eval_sym(symbol y)
{
  if (y.s[0]=='\'')
    dpush(sym2cell(y.s[1..$],y.c));
  else
  {
    header *h;
    if (y.c)
      h=y.c;
    else
      h=dicthead;
    h=dict_find_typed(h,y.s);
    if (!h)
    {
      seval("ds");
      if (y.c)
        h=y.c;
      else
        h=dicthead;
      writef("symbol:",y.s,"@",h,"\n");
      dict_show(h);
      h=dict_find_typed_debug(h,y.s);
    }
    eval_hdr(h);
  }
}

void eval_sub(cell c)
// evaluate
{
//  writefln("\n[eval ",cell_vs(c),"]");
  try
  {
    switch (c.th)
    {
      case CT_SYM : {eval_sym(cell2sym(c));break;}
      case CT_NFP : {cell2nfp(c)();break;}
      default     : dpush(c);
    }
  }
  catch (Object o)
  {
//    writefln("\n");
//    writefln("\np_load_module=",&p_load_module);
//    dict_show();
    writefln("\neval_sub: error (",o,") evaluating ",cell_vs(c));
    throw o;
  }
}

//void eval(void *c) {eval(cast(header*)c);}

void eval(cell c)
{
  try
  {
    if (c.th==CT_LST)
    {
      cell[] l=cell2lst(c);
      for (int k=0;k<l.length;k++)
        eval_sub(l[k]);
    }
    else
    {
      eval_sub(c);
    }
  }
  catch (Object o)
  {
    writefln("\neval: error (",o,") evaluating ",cell_vs(c));
  }
}
/*
void eval(cell c) // inlined eval-sub
{
  int k;
  try
  {
    if (c.th==CT_LST)
    {
      cell[] l=cell2lst(c);
      for (k=0;k<l.length;k++)
      {
        switch (l[k].th)
        {
          case CT_SYM : {eval_sym(cell2sym(cast(cell)(l[k])));break;}
          case CT_NFP : {cell2nfp(cast(cell)(l[k]))();break;}
          default     : dpush(l[k]);
        }
      }
    }
    else
    {
      switch (c.th)
      {
        case CT_SYM : {eval_sym(cell2sym(c));break;}
        case CT_NFP : {cell2nfp(c)();break;}
        default     : dpush(c);
      }
    }
  }
  catch (Object o)
  {
    writefln("\nerror (",o,") evaluating ",cell_vs(c));
  }
}
*/
int squote_int(char[] token)
{
  auto r=RegExp("^[0-9]+$");
  if (r.test(token))
  {
    dpush(int2cell(token.toInt()));
    return 1;
  }
  return 0;
}

int squote_flt(char[] token)
{
  auto s0="[+-]?(([0-9]+)|([0-9]*[.][0-9]+)|([0-9]+[.][0-9]*))";
  auto s1="^"~s0~"([eE]"~s0~")?$";
  auto r=RegExp(s1);
  if (r.test(token))
  {
    dpush(flt2cell(token.toFloat()));
    return 1;
  }
  return 0;
}

int squote_str(char[] token)
{
  auto r=RegExp("^\"(.|\n)*\"$");
  if (r.test(token))
//  if (token[0]=='"')
  {
//    printf(":%s:\n",asciiz(token));
    dpush(str2cell(token[1..$-1]));
    return 1;
  }
  return 0;
}

int squote_chr(char[] token)
{
  auto r=RegExp(r"^((#\d+)|(#\\.))$");
  if (r.test(token))
//  if (token[0]=='"')
  {
//    printf(":%s:\n",asciiz(token));
    if (token[1]=='\\')
    {
      dpush_chr(token[2]);
//      writefln("(",cast(int)token[2],")");
      return 1;
    }
    else
    {
      try
      {
        token=token[1..$];
        int i=token.toInt();
//        writefln("(",i,")");
        dpush(chr2cell(i));
        return 1;
      }
      catch (Object o)
      {
        writefln("\n*** ",o);
        return 0;
      }
    }
  }
  return 0;
}

int squote_sym(char[] token)
{
  if (token[0]=='&')
  {
    dpush(sym2cell(token[1..$],cast(header*)0));
    return 1;
  }
  if (token[0]=='\'')
  {
    dpush(sym2cell(token,dicthead));
    return 1;
  }
  header *h=dict_find_untyped(dicthead,token);
  if (!h)
  {
    writefln("undefined symbol: <",token,">");
    throw new Exception("squote_sym: undefined symbol");
  }
  if (hdr_get_flg(h)&FLG_M)
    eval_hdr(h);
  else
  {
//    writefln("symbol: ",token);
//    dpush(sym2cell(token,dicthead));
    dpush(sym2cell(token,h)); // to shorten search time
  }
  return 1;
}

void squote(char[] token)
// source string quote
{
//  printf("squote(%s)\n",asciiz(token));
  if (!token[0]) return; // code zero denotes end of source
  if (squote_int(token)) return;
  if (squote_flt(token)) return;
  if (squote_str(token)) return;
  if (squote_chr(token)) return;
  if (squote_sym(token)) return;
  throw new Exception(swritef("seval: unknown symbol <%s>",token));
}

void seval(char[] token)
// source string eval
{
  dpush(lst2cell(new_lst()));
  denter();
  squote(token);
  dleave();
//  if (token=="def")
//    seval("ds");
  eval(dpop());
//  ds_show();
}

void run()
{
  char[] token;
  while (src_left())
  {
    token=call_scan();
//    printf("[%s]\n",asciiz(token));
    seval(token);
  }
}
