
module grace_prim_cdst;

import grace_def;
import grace_cstk;
import grace_dict;
import grace_eval;
import grace_reader;
import grace_prim;

void reg_prims_cdst()
{
  // reader/scanner
  dict_addnf("peek"   ,&p_peek        ,           [],["chr"]);
  dict_addnf("read"   ,&p_read        ,           [],["chr"]);
  dict_addnf("scan"   ,&p_scan        ,           [],["str"]);
  dict_addnf("unread" ,&p_unread_chr  ,      ["chr"],[]);
  dict_addnf("unread" ,&p_unread_str  ,      ["str"],[]);
  dict_addnf("white?" ,&p_whitep      ,      ["chr"],[]);
  // structure
  dict_addnm("/*"     ,&p_slash_star  ,           [],[]);
  dict_addnm("//"     ,&p_slash_slash ,           [],[]);
  dict_addnm("["      ,&p_left_bracket,           [],["lst"]);
  dict_addnf("eval"   ,&p_eval        ,      [ "*" ],[]);
  dict_addnf("quote"  ,&p_quote       ,           [],[]);
  dict_addnf("def"    ,&p_def         ,[ "*" ,"sym"],[]);
  dict_addnf("defun"  ,&p_defun       ,[ "*" ,"sym"],[]);
  dict_addnf("defmac" ,&p_defmac      ,[ "*" ,"sym"],[]);
//  dict_addnf("setvar" ,&p_setvar      ,[ "*" ,"sym"],[]);
//  dict_addnf("setfun" ,&p_setfun      ,[ "*" ,"sym"],[]);
  dict_addnf("set"    ,&p_set         ,[ "*" ,"sym"],[]);
  dict_addnf("get"    ,&p_get         ,[      "sym"],[ "*" ]);
//  dict_addnf("setmac" ,&p_setmac      ,[ "*" ,"sym"],[]);
  dict_addnf("dict"   ,&p_dict        ,           [],[]);
  dict_addnf("type?"  ,&p_typep       ,           [],["str"]);
}

static this()
{
  reg_prims~=&reg_prims_cdst;
}

// --------------------
// ----- reader/scanner

void p_read()
{
  dpush_chr(src_read());
}

void p_peek()
{
  dpush_chr(src_peek());
}

void p_scan()
{
  dpush_str(src_scan());
}

void p_unread_str()
{
  src_unread(dpop_str());
}

void p_unread_chr()
{
  src_unread(dpop_chr());
}

void p_whitep()
{
  char c=dpop_chr();
  dpush_log(iswhite(c));
}

// --------------------
// ----- structuring

void p_left_bracket()
{
  char[] token;
  cell[] l;
  cell c;
  l=new_lst();
  c=lst2cell(l);
  dpush(c);
  denter();
  while (1)
  {
    token=call_scan();
    if (token=="]") break;
    squote(token);
  }
  dleave();
}

void p_slash_star()
{
  char[] s;
  while (1)
  {
    s=call_scan();
    if (s=="*/") break;
  }
  squote(call_scan());
}

void p_slash_slash()
{
  char c;
  char[] s;
//  writef("//");
  while (1)
  {
    c=cast(char)call_read();
    if (c=='\n') break;
//    writef(c);
  }
  s=call_scan();
//  writefln(" -> %s",s);
  squote(s);
}

void p_eval()
{
  cell c=dpop();
  eval(c);
}

void p_quote()
{
  squote(dpop_str());
}

void p_def()
{
//  writef("p_def ");
  cell n=dpop();
  cell d=dpop();
//  writef(cell2sym(n).s,":=",cell_vs(d),"\n");
  dict_add(cell2sym(n).s,d);
}

void p_defun()
{
  cell n=dpop();
  cell d=dpop();
  dict_addf(cell2sym(n).s,d);
}

void p_defmac()
{
  cell n=dpop();
  cell d=dpop();
  dict_addm(cell2sym(n).s,d);
}

void p_setvar()
{
  char[] n=dpop_sym().s;
  cell   d=dpop();
  int check(header *h)
  {
    if (hdr_get_nam(h)==n)
      if ((hdr_get_flg(h)&(FLG_F))==0)
        return 1;
    return 0;
  }
  header *h=dict_find(dicthead,&check);
  if (h)
    hdr_set_dat(h,d);
}

void p_setfun()
{
  char[] n=dpop_sym().s;
  cell   d=dpop();
  int check(header *h)
  {
    if (hdr_get_nam(h)==n)
      if ((hdr_get_flg(h)&(FLG_F|FLG_M))==FLG_F)
        return 1;
    return 0;
  }
  header *h=dict_find(dicthead,&check);
  if (h)
    hdr_set_dat(h,d);
}

void p_set()
{
  char[] n=dpop_sym().s;
  cell   d=dpop();
  int check(header *h)
  {
    if (hdr_get_nam(h)==n)
      return 1;
    return 0;
  }
  header *h=dict_find(dicthead,&check);
  if (h)
    hdr_set_dat(h,d);
}

void p_get()
{
  char[] n=dpop_sym().s;
  int check(header *h)
  {
    if (hdr_get_nam(h)==n)
      return 1;
    return 0;
  }
  header *h=dict_find(dicthead,&check);
  if (h)
    dpush(hdr_get_dat(h));
}

void p_setmac()
{
  char[] n=dpop_sym().s;
  cell   d=dpop();
  int check(header *h)
  {
    if (hdr_get_nam(h)==n)
      if ((hdr_get_flg(h)&(FLG_M))==FLG_M)
        return 1;
    return 0;
  }
  header *h=dict_find(dicthead,&check);
  if (h)
    hdr_set_dat(h,d);
}

void p_dict()
{
  cell[] l=dict_aslist();
  dpush_lst(l);
}

void p_typep()
{
  cell c=dpop();
  dpush_str(uh2str(c.th));
}

