
module grace_prim_stack;

import grace_def;
import grace_cstk;
import grace_dict;
import grace_eval;
import grace_reader;
import grace_prim;

void reg_prims_stack()
{
  // --- stack manipulation
  // ... -
  dict_addnf("zapp",&p_zapp,[],[]);
  // a b - a b b
  dict_addnf("dup"   ,&p_dup       ,            [ "*" ],      [ "*" , "*" ]);
  // a b - a
  dict_addnf("drop"  ,&p_drop      ,            [ "*" ],                 []);
  // a b - b a
  dict_addnf("swap"  ,&p_swap      ,      [ "*" , "*" ],      [ "*" , "*" ]);
  dict_addnf("extent",&p_extent    ,                 [],            ["int"]);
  // a [ B ] - B
  dict_addnf("enter" ,&p_enter     ,            ["lst"],                 []);
  //
  dict_addnf("leave" ,&p_leave     ,                 [],            ["lst"]);
  //
  dict_addnf("entern",&p_enter_new ,                 [],                 []);
  // a b - a b a
  dict_addnf("over"  ,&p_over      ,      [ "*" , "*" ],[ "*" , "*" , "*" ]);
  // a b - b
  dict_addnf("nip"   ,&p_nip       ,[       "*" , "*" ],[             "*" ]);
  // a b - b a b
  dict_addnf("tuck"  ,&p_tuck      ,[       "*" , "*" ],[ "*" , "*" , "*" ]);
  // a b c - b c a
  dict_addnf("rot"   ,&p_rot       ,[ "*" , "*" , "*" ],[ "*" , "*" , "*" ]);
  // a b c - c a b
  dict_addnf("-rot"  ,&p_mrot      ,[ "*" , "*" , "*" ],[ "*" , "*" , "*" ]);
  // dive / soar
  dict_addnf("dive"  ,&p_dive,[],[]);
  dict_addnf("soar"  ,&p_soar,[],[]);
  dict_addnf("ndive" ,&p_ndive,["int"],[]);
  dict_addnf("nsoar" ,&p_nsoar,["int"],[]);
  dict_addnf("xdive" ,&p_xdive,[],[]);
  dict_addnf("xsoar" ,&p_xsoar,[],[]);
  //
  dict_addnf("top"   ,&p_top,["*"],[]);
  // depth
  dict_addnf("depth" ,&p_depth     ,                 [],            ["int"]);
  // height
  dict_addnf("height",&p_height    ,                 [],            ["int"]);
  // extent
}

static this()
{
  reg_prims~=&reg_prims_stack;
}
// --------------------
// ----- stack

cell[] secs;

void p_top()
{
  stack_top();
}

void p_zapp()
{
  stack_zapp();
}

void p_dive()
{
  dive();
}

void p_soar()
{
  soar();
}

void p_ndive()
{
  int k=dpop_int();
  while (k--)
    dive();
}

void p_nsoar()
{
  int k=dpop_int();
  while (k--)
    soar();
}

void p_xdive()
{
  xdive();
}

void p_xsoar()
{
  xsoar();
}

void p_dup()
{
  dpush(dtos());
}

void p_drop()
{
  dpop();
}

void p_swap()
{
  cell c1=dpop();
  cell c2=dtos();
  dtoset(c1);
  dpush(c2);
}

void p_depth()
{
  dpush(int2cell(stk.length));
}

void p_height()
{
  dpush(int2cell(secs.length));
}

void p_extent()
{
  p_depth();
  p_height();
  dpush_int(dpop_int()+dpop_int());
}

void p_enter_new()
{
  dpush(lst2cell(new_lst()));
  denter();
}

void p_enter()
{
  denter();
}

void p_leave()
{
  xsoar();
  dleave();
}

void p_over()
{
  cell c2=dpop();
  cell c1=dtos();
  dpush(c2);
  dpush(c1);
}

void p_nip()
{
  cell c2=dpop();
  cell c1=dpop();
  dpush(c2);
}

void p_tuck()
{
  cell c2=dpop();
  cell c1=dpop();
  dpush(c2);
  dpush(c1);
  dpush(c2);
}

void p_rot()
{
  cell c3=dpop();
  cell c2=dpop();
  cell c1=dpop();
  dpush(c2);
  dpush(c3);
  dpush(c1);
}

void p_mrot()
{
  cell c3=dpop();
  cell c2=dpop();
  cell c1=dpop();
  dpush(c3);
  dpush(c1);
  dpush(c2);
}

