
module grace_reader;

import grace_def;
import grace_cstk;
import grace_dict;
import grace_eval;

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// -------------------- source handling

char[][] srcstk;
int[] srcpos;

int src_left()
{
  return srcstk.length;
}

void src_push_str(char[] str)
{
  srcstk~=str;
  srcpos~=0;
}

void src_push_file(char[] fn)
{
  FILE *f;
  void[] buf;
  buf=read(fn);
  src_unread(cast(char[])buf);
}

void src_show()
{
  int k;
  for (k=srcstk.length;k;)
  {
    --k;
    writef(srcstk[k][srcpos[k]..$]);
  }
}

char src_read()
{
  if (srcstk.length)
  { // strings left of stack
    if (srcstk[$-1].length>srcpos[$-1])
    { // chars left in string
      return srcstk[$-1][srcpos[$-1]++];
    }
    else
    { // current string empty
      srcstk.length=srcstk.length-1;
      srcpos.length=srcpos.length-1;
      return src_read();
    }
  }
  else
  { // no more strings
    return 0;
  }
}

char src_peek()
{
  if (srcstk.length)
  { // strings left of stack
    if (srcstk[$-1].length>srcpos[$-1])
    { // chars left in string
      return srcstk[$-1][srcpos[$-1]];
    }
    else
    { // current string empty
      srcstk.length=srcstk.length-1;
      srcpos.length=srcpos.length-1;
      return src_peek();
    }
  }
  else
  { // no more strings
    return 0;
  }
}

void src_unread(char c)
{
  char[] s;
  if (srcstk.length)
  { // strings left of stack
    if (srcpos[$-1]>0)
    { // chars left before position
      srcstk[$-1][--srcpos[$-1]]=c;
    }
    else
    { // create string
      s.length=0x100;
      s[$-1]=c;
      srcstk~=s;
      srcpos~=0xff;
    }
  }
  else
  { // create string
    s.length=0x100;
    s[$-1]=c;
    srcstk~=s;
    srcpos~=0xff;
  }
}

alias src_push_str src_unread;

char[] src_scan()
{
  char[] s;
  char   c;
  // strip leading whitespace
  do {
    c=call_read();
  } while (iswhite(c));
  // string ?
  if (c=='"')
  {
    for (;;)
    {
      s~=c;
      c=call_read();
      if (c=='"')
        return (s~c);
      if (c=='\\')
      {
        c=call_read();
        if (c=='n') c='\n';
      }
    }
  }
  // collect token
	for (;;)
  {
    if ((c=='[')||(c==']'))
    {
      if (s.length)
      {
        call_unread(c);
        return s;
      }
      else
      {
        return (s~c);
      }
    }
		s~=c;
		c=call_read();
		// end of source ?
		if (!c)
      return s;
    // end of token ?
		if (iswhite(c))
		{ // unread whitespace (needed by // macro)
      call_unread(c);
      return s;
    }
	}
}

void src_tst()
{
	char[] s;
  char c;
  do {
    writefln(src_scan());
  } while (src_left());
  src_push_str("abc ");
  src_push_str("123 ");
  src_unread(' ');
  src_unread('!');
  c=' ';
  while (c)
  {
    c=src_read();
    writef(c);
  }
  writef("\n");
}

char call_read()
{
  seval("read");
  return dpop_chr();
}

char call_peek()
{
  seval("peek");
  return dpop_chr();
}

void call_unread(char c)
{
  dpush_chr(c);
  seval("unread");
}

void call_unread(char[] s)
{
  dpush_str(s);
  seval("unread");
}

char[] call_scan()
{
  seval("scan");
  return dpop_str();
}

